"""Apply a vocab mapping converting a list of tokens in a column
into a list of integers"""

from abc import abstractmethod
from math import ceil
from sys import getsizeof
from typing import List, Union

import numpy as np
import polars as pl
import pyarrow as pa

from bardi.data.utils.pyarrow_utils import chunk_pyarrow_table
from bardi.nlp_engineering.utils.validations import (
    validate_list_str_cols,
    validate_pyarrow_table,
)
from bardi.pipeline import DataWriteConfig, Step


class PostProcessor(Step):
    """The post processor maps a vocab to a list of tokens

    Avoid the direct instantiation of the PostProcessor class
    and instead instantiate one of the child classes depending
    on hardware configuration
    """

    def __init__(
        self,
        fields: Union[str, List[str]],
        field_rename: str = None,
        id_to_token: dict = None,
        concat_fields: bool = False,
    ):
        """instantiate a post processor object"""
        self.fields = fields
        if isinstance(fields, str):
            self.fields = [fields]

        self.field_rename = "X"
        if field_rename:
            self.field_rename = field_rename

        self.mapping = None
        if id_to_token:
            # the actual application of the mapping requires
            # flipping the keys and values to have {token: id}
            self.mapping = {token: int(id) for id, token in id_to_token.items()}

        self.concat_fields = concat_fields
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }  # Default write config

    @abstractmethod
    def run(self):
        """to be implemented in child class"""
        pass


class CPUPostProcessor(PostProcessor):
    """The post processor maps a vocab to a list of tokens

    Implementation of the PostProcessor for CPU computation

    Attributes:
        fields : Union[str, List[str]],
            the name of the column containing a list of tokens
            that will be mapped to integers using a vocab
        field_rename : str
            optional ability to rename the supplied field with
            the field_rename value
        id_to_token : dict
            optional vocabulary in the form of {id: token} that
            will be used to map the tokens to integers. This
            is optional for the construction of the object, and
            can alternatively be provided in the run method.
            This flexibility handles the use of a pre-existing
            vocab versus creating a vocab during a pipeline run.
        concat_fields : bool
            indicate if you would like for fields to be concatenated
            into a single column or left as separate columns

    Methods:
        run : run the step's primary function
        get_parameters : get a dictionary representation of the step object
        set_write_config : Alter the default file writing configuration
        write_outputs : Write output data to a file
    """

    def __init__(self, *args, **kwargs):
        """
        Keyword Arguments:
            fields : Union[str, List[str]],
                the name of the column containing a list of tokens
                that will be mapped to integers using a vocab
            field_rename : str
                optional ability to rename the supplied field with
                the field_rename value
            id_to_token : dict
                optional vocabulary in the form of {id: token} that
                will be used to map the tokens to integers. This
                is optional for the construction of the object, and
                can alternatively be provided in the run method.
                This flexibility handles the use of a pre-existing
                vocab versus creating a vocab during a pipeline run.
            concat_fields : bool
                indicate if you would like for fields to be concatenated
                into a single column or left as separate columns"""
        super().__init__(*args, **kwargs)

    def run(
        self,
        data: pa.Table,
        artifacts: dict = None,
        id_to_token: dict = None,
    ) -> pa.Table:
        """Run a post processor using CPU computation

        The post processor relies on receiving a vocab to map. The
        vocab can be supplied in multiple ways:
          - id_to_token at object creation
          - contained in the pipeline artifacts dictionary passed
            to the run method referenced by the key, 'id_to_token'
          - id_to_token in the run method

        Keyword Arguments:
            data : PyArrow Table
                The data to be processed. The data must contain the
                column specified by field at object creation
            artifacts : dict
                A dictionary of pipeline artifacts which contains
                a vocab referenced by the key, 'id_to_token'
            id_to_token : dict
                If a vocab wasn't passed at object creation or
                through the pipeline artifacts dict, then it
                must be passed here as a final option

        Returns:
            Tuple(PyArrow Table, dict)
                A tuple with:
                 - the first element holding a PyArrow Table of data
                 processed with the post processor
                 - the second element of the tuple intended for
                 artifacts is None

        Raises
            AttributeError
                The vocab (id_to_token) wasn't supplied
                either at object creation or to the run method
            TypeError
                The run method was not supplied a PyArrow Table
        """

        # Perform validations
        validate_pyarrow_table(data=data)
        validate_list_str_cols(fields=self.fields, data=data)

        # Check if vocab was passed through the artifacts dict
        if "id_to_token" in artifacts.keys():
            id_to_token = artifacts["id_to_token"]

        # Check if a vocab was passed directly to the method -
        # the route used if a vocab was created during script execution
        if id_to_token:
            self.mapping = {token: id for id, token in id_to_token.items()}
        # Check if a vocab was referenced in the initial object creation -
        # the route used if there was a pre-existing vocab
        elif not self.mapping:
            # If a vocab wasn't referenced in either place, then there
            # isn't a vocab to use in this run method
            raise AttributeError(
                "id_to_token must be provided to either "
                "the PostProcessor object instantiation "
                "or to the run method"
            )

        # Grab the <unk> id from the supplied vocab
        try:
            self.unk_id = self.mapping["<unk>"]
        except KeyError:
            self.unk_id = None  # If there isn't an unk id

        # Map tokens to ids using the supplied vocab.
        # If a token encountered isn't in the vocab, it defaults to
        # the <unk> value. The field passed is renamed to 'X'

        # This step is memory intensive, so the data is split into
        # 128 MB chunks and operated upon to avoid OoM errors
        data_size = getsizeof(data) * 1e-6
        chunks = ceil(data_size / 128)
        column_order = data.column_names
        tables = chunk_pyarrow_table(
            data=data, row_count=data.num_rows, min_batches=chunks
        )

        self.mapping[None] = None

        for i, table in enumerate(tables):
            chunk_df = pl.from_arrow(table)
            # create a unique id for each row in the chunk
            # this is needed for exploding and grouping
            # the data for efficinet execution
            index_list = np.arange(0, chunk_df.height)
            index_col = pl.Series(index_list)
            chunk_df = chunk_df.with_columns(index_col.alias("unique_id"))

            for field in self.fields:
                meta_df = chunk_df.drop(field)
                tmp_df = (
                    chunk_df.select(["unique_id", field])
                    .explode(field)
                    .with_columns(
                        pl.col(field).map_dict(
                            self.mapping, default=self.unk_id, return_dtype=pl.Int64()
                        )
                    )
                    .group_by("unique_id", maintain_order=True)
                    .agg(pl.col(field))
                )

                chunk_df = meta_df.join(tmp_df, on="unique_id", how="left")
            if i == 0:
                df = chunk_df.select(*column_order)
            else:
                df = df.vstack(chunk_df.select(*column_order))

        # concat the fields if desired
        if self.concat_fields and len(self.fields) > 1:
            # Concatenate the values in the fields.
            # Fields at this point are assumed to be large lists

            current_field = self.fields[0]
            for i in range(len(self.fields) - 1):
                next_field = self.fields[i + 1]
                df = df.with_columns(
                    pl.col(current_field)
                    .list.concat(next_field)
                    .list.drop_nulls()
                    .alias("X")
                ).drop(next_field)
                current_field = "X"
            df = df.drop(self.fields[0])

            # concat field is called 'X', rename if
            # other name provided, defaults to 'X'
            df = df.rename({"X": self.field_rename})

        # if concat was not run and a single field was provided
        # the field can be renamed
        elif len(self.fields) == 1:
            df = df.rename({self.fields[0]: self.field_rename})

        data = df.to_arrow()

        return (data, None)

    def get_parameters(self):
        """Retrive the post-processor object configuration
        Does not return the mapping (vocab) as it can be large

        Returns:
            a dictionary representation of the post-processor's attributes
        """
        params = vars(self).copy()
        params.pop("mapping")
        return params



"""Train a Word2Vec model and create a vocab and word embeddings"""

import os
from abc import abstractmethod
from multiprocessing import cpu_count
from os.path import isfile
from typing import List, TypedDict, Union, Tuple

import numpy as np
import polars as pl
import pyarrow as pa
from gensim.utils import RULE_DEFAULT, RULE_DISCARD
from gensim.models import Word2Vec

from bardi.data import data_handlers
from bardi.nlp_engineering.utils.validations import (
    validate_pyarrow_table,
    validate_list_str_cols,
)
from bardi.pipeline import DataWriteConfig, Step


class EmbeddingGeneratorArtifactsWriteConfig(TypedDict):
    """Indicates the keys and data types expected in an
    artifacts write config dict for the embedding generator
    if overwriting the default configuration."""

    vocab_format: str
    vocab_format_args: dict
    embedding_matrix_format: str
    embedding_matrix_format_args: dict


class EmbeddingGenerator(Step):
    """The embedding generator provides interface to create
    word embeddings or vector representations of words (tokens).

    The embedding generator uses Word2Vec model from Gensim library.

    Avoid the direct instantiation of the PreTokenizer class
    and instead instantiate one of the child classes depending
    on hardware configuration.
    """

    def __init__(
        self,
        fields: Union[str, List[str]],
        load_saved_model: bool = False,
        checkpoint_path: str = None,
        cores: int = cpu_count(),
        min_word_count: int = 10,
        window: int = 5,
        vector_size: int = 300,
        sample: float = 6e-5,
        min_alpha: float = 0.007,
        negative: int = 20,
        epochs: int = 30,
        seed: int = 42,
        vocab_exclude_list: List[str] = [],
    ):
        """instantiate an embedding generator object"""
        if isinstance(fields, str):
            self.fields = [fields]
        else:
            self.fields = fields
        self.load_saved_model = load_saved_model
        self.checkpoint_path = checkpoint_path
        self.cores = cores
        self.min_word_count = min_word_count
        self.window = window
        self.vector_size = vector_size
        self.sample = sample
        self.min_alpha = min_alpha
        self.negative = negative
        self.epochs = epochs
        self.seed = seed
        self.vocab_exclude_list = vocab_exclude_list
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }
        self._artifacts_write_config: EmbeddingGeneratorArtifactsWriteConfig = {
            "vocab_format": "json",
            "vocab_format_args": {},
            "embedding_matrix_format": "npy",
            "embedding_matrix_format_args": {},
        }

    def set_write_config(
        self,
        data_config: DataWriteConfig = None,
        artifacts_config: EmbeddingGeneratorArtifactsWriteConfig = None,
    ):
        """Overwrite the default file writing configurations"""

        if data_config:
            self._data_write_config = data_config
        if artifacts_config:
            self._artifacts_write_config = artifacts_config

    @abstractmethod
    def run(self):
        """to be implemented in child class"""
        pass


class CPUEmbeddingGenerator(EmbeddingGenerator):
    """The embedding generator provides interface to create
    word embeddings or vector representations of words (tokens).

    The embedding generator uses Word2Vec model from Gensim library.

    EmbeddingGenerator specific for CPU computation.

    Attributes:
        fields : Union[str, List[str]]
            the name of the column(s) containing text to be considered
            in the vocab and used in Word2Vec
        load_saved_model : bool
            if True, use pre-trained word2vec model.
        checkpoint_path : str
            path to word2vec model checkpoint.
        cores : int
            numbers of cores to run word2vec model on.
        min_word_count : int
            ignores all words with total frequency lower than this.
        window : int
            maximum distance between the current and predicted word.
        vector_size : int
            output embedding size.
        sample : float
            the threshold for configuring which high-frequency
            words are randomly ownsamples, use range (0, 1e-5).
        min_alpha : float
            learning rate will linearly drop to min_alpha
            as training progresses.
        negative : int
            if > 0, negative sampling will be used.
        epochs : int
            total number of iterations of all training data in
            the training of the word2vec model
        seed : int
            seed for random number generetor, for deterministic
            run you need thread = 1 (aka cpu core)
            and PYTHONHASHSEED
        vocab_exclude_list : List[str]
            provide a list of tokens that may be present in the
            text that you would like to exclude from the vocab and
            from Word2Vec

    Methods:
        run : run the step's primary function
        get_parameters : get a dictionary representation of the step object
        set_write_config : alter the default file writing configuration
        write_outputs : Write output data to a file
    """

    def __init__(self, *args, **kwargs):
        """Create an object of the CPUEmbeddingGenerator Classs

        Keyword Arguments:
            fields : Union[str, List[str]]
                the name of the column(s) containing text to be considered
                in the vocab and used in Word2Vec
            load_saved_model : bool
                if True, use pre-trained word2vec model.
            checkpoint_path : str
                path to word2vec model checkpoint.
            cores : int
                numbers of cores to run word2vec model on.
            min_word_count : int
                ignores all words with total frequency lower than this.
            window : int
                maximum distance between the current and predicted word.
            vector_size : int
                output embedding size.
            sample : float
                the threshold for configuring which high-frequency
                words are randomly ownsamples, use range (0, 1e-5).
            min_alpha : float
                learning rate will linearly drop to min_alpha
                as training progresses.
            negative : int
                if > 0, negative sampling will be used.
            epochs : int
                total number of iterations of all training data in
                the training of the word2vec model
            seed : int
                seed for random number generetor, for deterministic
                run you need thread = 1 (aka cpu core)
                and PYTHONHASHSEED
            vocab_exclude_list : List[str]
                provide a list of tokens that may be present in the
                text that you would like to exclude from the vocab and
                from Word2Vec
        """
        super().__init__(*args, **kwargs)

        self.w2v_model = None
        self.embedding_matrix = None
        self.id_to_token = None

    def run(self, data: pa.Table, artifacts: dict) -> Tuple[pa.Table, dict]:
        """Runs a CPU-based embedding generator method based on the
        configuration used to create the object of the CPUEmbeddingGenerator
        class

        Arguments:
            data : pyarrow.Table
                a pyarrow Table containing at least one list column containing
                text
            artifacts : dict
                artifacts are not consumed in this run method, but must be
                received to operate correctly in the pipeline run method

        Returns:
            Tuple[pyarrow.Table, dict]
                The first position is a pyarrow.Table of pre-tokenized data
                The second position is a dictionary of artifacts. The dict will
                contain keys for "embedding_matrix" and "id_to_token".
        """
        # Perform validations
        validate_pyarrow_table(data=data)
        validate_list_str_cols(fields=self.fields, data=data)

        # Use the Polars library to convert a text column into a list of lists
        # [each list here is a set of tokens]. The list of lists is called
        # df_words_lists. If there are multiple text columns, the df_words_list
        # is extended with contents of rows of the additional columns.
        text_data = pl.from_arrow(data.select(self.fields))
        df_words_lists = []
        for field in self.fields:
            df_words_lists.extend(text_data[field].drop_nulls().to_list())

        if self.load_saved_model and isfile(self.checkpoint_path):
            self.w2v_model = Word2Vec.load(self.checkpoint_path)
        else:
            # Configure Word2Vec
            self.w2v_model = Word2Vec(
                min_count=self.min_word_count,
                window=self.window,
                vector_size=self.vector_size,
                sample=self.sample,
                min_alpha=self.min_alpha,
                negative=self.negative,
                workers=self.cores,
                seed=self.seed,
            )

            # Word2Vec vocabulary trim rule
            def vocab_trim_rule(word: str, count: int, min_count: int):
                if count < min_count:
                    return RULE_DISCARD
                elif word in self.vocab_exclude_list:
                    return RULE_DISCARD
                else:
                    return RULE_DEFAULT

            # Train Word2Vec Model
            self.w2v_model.build_vocab(df_words_lists, trim_rule=vocab_trim_rule)
            self.w2v_model.train(
                df_words_lists,
                total_examples=self.w2v_model.corpus_count,
                epochs=self.epochs,
            )
        # If a checkpoint path was set, the word2vec model will be saved
        if self.checkpoint_path:
            self.w2v_model.save(self.checkpoint_path)

        # Sort the vocabulary
        sorted_tokens = sorted(self.w2v_model.wv.index_to_key)
        sorted_embeddings = []

        for token in sorted_tokens:
            sorted_embeddings.append(
                self.w2v_model.wv.vectors[self.w2v_model.wv.key_to_index[token]]
            )

        # Mapping between an index in the embedding matrix and a token
        self.id_to_token = {
            token_id: token
            for token_id, token in enumerate(["<pad>"] + sorted_tokens + ["<unk>"])
        }

        # Create embedding matrix with embeddings for <pad> and <unk> tokens
        pad_token_emb = np.zeros((1, self.vector_size))
        unkn_token_emb = np.random.random((1, self.vector_size)) * 0.01
        self.embedding_matrix = np.append(
            np.append(pad_token_emb, sorted_embeddings, axis=0), unkn_token_emb, axis=0
        )

        # Set up the artifacts dict to return
        produced_artifacts = {
            "embedding_matrix": self.embedding_matrix,
            "id_to_token": self.id_to_token,
        }

        return (data, produced_artifacts)

    def write_artifacts(self, write_path: str, artifacts: Union[dict, None]) -> None:
        """Write the oartifactsproduced by the embedding_generator

        Keyword Arguments:
            write_path : str
                Path is a directory where files will be written
            artifacts : Union[dict, None]
                Artifacts is a dictionary of artifacts produced in this step.
                Expected keys are: "id_to_token" and "embedding_matrix"

        Returns:
            None
        """
        id_to_token = artifacts["id_to_token"]
        # Add the filename with the appropriate filetype to the write path
        id_to_token_path = os.path.join(
            write_path,
            (f"id_to_token" f'.{self._artifacts_write_config["vocab_format"]}'),
        )
        # Call the data handler write_file function
        data_handlers.write_file(
            data=id_to_token,
            path=id_to_token_path,
            format=self._artifacts_write_config["vocab_format"],
            **self._artifacts_write_config["vocab_format_args"],
        )

        embedding_matrix = artifacts["embedding_matrix"]
        # Add the filename with the appropriate filetype to the write path
        embedding_matrix_path = os.path.join(
            write_path,
            (
                f"embedding_matrix"
                f'.{self._artifacts_write_config["embedding_matrix_format"]}'
            ),
        )
        # Call the data handler write_file function
        data_handlers.write_file(
            data=embedding_matrix,
            path=embedding_matrix_path,
            format=self._artifacts_write_config["embedding_matrix_format"],
            **self._artifacts_write_config["embedding_matrix_format_args"],
        )

    def get_parameters(self) -> dict:
        """Retrive the embedding generetor object configuration

        Returns:
            a dictionary representation
            of the embedding generetor object's attributes
        """

        params = vars(self).copy()
        params["vocab_size"] = len(params["id_to_token"].keys())
        params.pop("id_to_token")
        params.pop("embedding_matrix")
        params["w2v_model"] = str(type(self.w2v_model))

        return params



"""Split text columns into lists of tokens using simple patterns"""

from abc import abstractmethod
from typing import List, Tuple, Union

import polars as pl
import pyarrow as pa

from bardi.nlp_engineering.utils.validations import (
    validate_pyarrow_table,
    validate_str_cols,
)
from bardi.pipeline import Step


class PreTokenizer(Step):
    """The pre-tokenizer breaks down text into smaller units
    before further tokenization is applied.

    Avoid the direct instantiation of the PreTokenizer class
    and instead instantiate one of the child classes depending
    on hardware configuration.
    """

    def __init__(self, fields: Union[str, List[str]], split_pattern: str = " "):
        """instantiate a pre tokenizer object"""
        if isinstance(fields, str):
            self.fields = [fields]
        else:
            self.fields = fields
        self.split_pattern = split_pattern

    @abstractmethod
    def run(self):
        """to be implemented in child class"""
        pass


class CPUPreTokenizer(PreTokenizer):
    """The pre tokenizer breaks down text into smaller units
    before further tokenization is applied.

    Implementation of the PreTokenizer for CPU computation

    Attributes:
        fields : Union[str, List[str]],
            the name of the column(s) containing text
        split_pattern : str
            a specific pattern of characters used to divide a string
            into smaller segments or tokens.
            By default, the split is done on a single space character.

    Methods:
        run : run the step's primary function
        get_parameters : get a dictionary representation of the step object
        set_write_config : Alter the default file writing configuration
        write_outputs : Write output data to a file
    """

    def __init__(self, *args, **kwargs):
        """
        Keyword Arguments:
            fields : Union[str, List[str]],
                the name of the column(s) containing text
            split_pattern : str
                a specific pattern of characters used to divide a string
                into smaller segments or tokens.
                By default, the split is done on a single space character.
        """
        super().__init__(*args, **kwargs)

    def run(
        self, data: pa.Table, artifacts: dict = None
    ) -> Tuple[pa.Table, Union[dict, None]]:
        """Runs a CPU-based pre-tokenizer method based on the configuration
        used to create the object of the CPUPreTokenizer class

        Arguments:
            data : pyarrow.Table
                a pyarrow Table containing at least one text column of type
                string or large_string
            artifacts : dict
                artifacts are not used in this run method, but must be received
                to operate correctly in the pipeline run method

        Returns:
            Tuple[pyarrow.Table, dict]
                The first position is a pyarrow.Table of pre-tokenized data
                The second position is a dictionary of artifacts. No artifacts
                are produced in this run method, so the second position will
                return None.
        """
        # Perform validations
        validate_pyarrow_table(data=data)
        validate_str_cols(fields=self.fields, data=data)

        # Split text fields into lists of tokens
        df = pl.from_arrow(data).with_columns(
            [
                (
                    pl.col(field)
                    .str.split(by=self.split_pattern)
                    .list.eval(pl.element().filter(pl.element() != ""))
                    .alias(field)
                )
                for field in self.fields
            ]
        )

        data = df.to_arrow()

        return (data, None)



import os
from abc import abstractmethod
from typing import List, Tuple, TypedDict, Union

import polars as pl
import pyarrow as pa

from bardi.data import data_handlers
from bardi.nlp_engineering.utils.validations import validate_pyarrow_table
from bardi.pipeline import DataWriteConfig, Step


class LabelProcessorArtifactsWriteConfig(TypedDict):
    """Indicates the keys and data types expected in an
    artifacts write config dict for the label processor
    if overwriting the default configuration."""

    id_to_label_format: str
    id_to_label_args: Union[dict, None]


class LabelProcessor(Step):
    """The label processor creates and maps a label vocab

    Avoid the direct instantiation of the LabelProcessor class
    and instead instantiate one of the child classes.
    """

    def __init__(self, fields: Union[str, List[str]], method: str = "unique"):
        """Create an object of the LabelProcessor class

        Keyword Arguments:
            fields : Union[str, List[str]]
                the name of a label column of which the values
                are used to generate a standardized mapping and
                then that mapping is applied to the column
            method : str
                currently only a default 'unique' method is
                supported which maps each unique value in the
                column to an id
        """
        self.fields = fields
        if isinstance(fields, str):
            self.fields = [fields]
        else:
            self.fields = fields
        self.method = method
        self.mapping = {}
        self.id_to_label = {}
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }
        self._artifacts_write_config: LabelProcessorArtifactsWriteConfig = {
            "id_to_label_format": "json",
            "id_to_label_format_args": {},
        }

    def set_write_config(
        self,
        data_config: DataWriteConfig = None,
        artifacts_config: LabelProcessorArtifactsWriteConfig = None,
    ):
        """Overwrite the default file writing configurations"""
        if data_config:
            self._data_write_config = data_config
        if artifacts_config:
            self._artifacts_write_config = artifacts_config

    @abstractmethod
    def run(self):
        """to be implemented in child class"""
        pass


class CPULabelProcessor(LabelProcessor):
    """The label processor creates and maps a label vocab

    Implementation of the LabelProcessor for CPU computation

    Attributes:
        fields : Union[str, List[str]]
            the name(s) of label column(s) of which the values
            are used to generate a standardized mapping and
            then that mapping is applied to the column(s)
        method : str
            currently only a default 'unique' method is
            supported which maps each unique value in the
            column to an id
        mapping : dict
            mapping dict of the form {label: id} used to convert
            labels in the column to ids
        id_to_label : dict
            the reverse of mapping. of the form {id: label} used
            downstream to map the ids back to the original label
            values

    Methods:
        run : run the step's primary function
        get_parameters : get a dictionary representation of the step object
        set_write_config : Alter the default file writing configuration
        write_outputs : Write output data to a file
    """

    def __init__(self, *args, **kwargs):
        """Create an object of the CPULabelProcessor class

        Keyword Arguments:
            fields : Union[str, List[str]]
                the name of a label column of which the values
                are used to generate a standardized mapping and
                then that mapping is applied to the column
            method : str
                currently only a default 'unique' method is
                supported which maps each unique value in the
                column to an id
        """
        super().__init__(*args, **kwargs)

    def run(self, data: pa.Table, artifacts: dict = None) -> Tuple[pa.Table, dict]:
        """Run a label processor using CPU computation

        Keyword Arguments:
            data : PyArrow Table
                The data to be processed. The data must contain the
                column specified by 'field' at object creation
            artifacts : dict
                artifacts are not used in this run method, but must be received
                to operate correctly in the pipeline run method

        Returns:
            Tuple[pyarrow.Table, dict]
                The first position is a pyarrow.Table of processed data
                The second position is a dictionary of artifacts. The dict will
                contain a key for "id_to_label".

        Raises
            NotImplementedError
                A value other than 'unique' was provided for the
                label processor's method
            TypeError
                The run method was not supplied a PyArrow Table
        """
        # Perform validations
        validate_pyarrow_table(data=data)

        # Confirm that the method is passed a PyArrow Table
        produced_artifacts = {}
        for field in self.fields:
            if self.method == "unique":
                try:
                    # Get the unique labels
                    vals = (
                        pl.from_arrow(data).get_column(field).unique().sort()
                    ).to_list()

                    # Create a mapping for the labels
                    field_mapping = {label: id for id, label in enumerate(vals)}
                    # id_to_label is flipped vs what is needed for mapping
                    field_id_to_label = {
                        str(id): str(label) for label, id in field_mapping.items()
                    }
                    self.id_to_label[field] = field_id_to_label

                    df = pl.from_arrow(data).with_columns(
                        pl.col(field).map_dict(field_mapping, default=None)
                    )

                except Exception as e:
                    print(
                        f"An error occured: {e} "
                        "Processing without creating the label "
                        f"mapping for {field}"
                    )
                    artifact = {"id_to_label": {field: None}}
                    return (data, artifact)

            else:
                raise NotImplementedError("Only unique mapping is currently supported")

            data = df.to_arrow()

        # Set up the artifacts dict to return
        produced_artifacts = {"id_to_label": self.id_to_label}

        return (data, produced_artifacts)

    def write_artifacts(self, write_path: str, artifacts: Union[dict, None]) -> None:
        """Write the outputs produced by the label_processor

        Keyword Arguments:
            write_path : str
                Path is a directory where files will be written
            artifacts : Union[dict, None]
                Artifacts is a dictionary of artifacts produced in this step.
                Expected key is: "id_to_label"

        Returns:
            None
        """
        id_to_token = artifacts["id_to_label"]
        # Add the filename with the appropriate filetype to the write path
        id_to_token_path = os.path.join(
            write_path,
            (f"id_to_label" f'.{self._artifacts_write_config["id_to_label_format"]}'),
        )
        # Call the data handler write_file function
        data_handlers.write_file(
            data=id_to_token,
            path=id_to_token_path,
            format=self._artifacts_write_config["id_to_label_format"],
            **self._artifacts_write_config["id_to_label_format_args"],
        )

    def get_parameters(self) -> dict:
        """Retrive the label processor object configuration

        Does not return the mapping (vocab), but does return the id_to_label
        dict. This is because the mapping is just the reverse of id_to_label.

        Returns:
            a dictionary representation of the splitter object's attributes
        """
        params = vars(self).copy()
        params.pop("mapping")
        params.pop("id_to_label")
        return params



"""Clean text with custom sets of regular expressions"""

from abc import abstractmethod
from typing import List, Tuple, Union

import polars as pl
import pyarrow as pa

from bardi.nlp_engineering.regex_library.regex_lib import RegexSubPair
from bardi.nlp_engineering.utils.validations import (validate_pyarrow_table,
                                                     validate_str_cols)
from bardi.pipeline import DataWriteConfig, Step


class Normalizer(Step):
    """Normalizer cleans and standardizes text input using regular expression
    substitutions. Lowercasing is also applied if desired.

    Avoid the direct instantiation of the PreTokenizer class
    and instead instantiate one of the child classes depending
    on hardware configuration.
    """
    def __init__(self,
                 fields: Union[str, List[str]],
                 regex_set: List[RegexSubPair],
                 lowercase: bool = True,
                 ):
        # Normalizer Configuration
        self.fields = fields
        self._data_write_config: DataWriteConfig = {
            "data_format": 'parquet',
            "data_format_args": {"compression": 'snappy',
                                 "use_dictionary": False}
        }  # Default write config
        self.lowercase = lowercase
        # List of regex substitutions to be applied
        self.regex_set: List[RegexSubPair] = regex_set

    @abstractmethod
    def run(self):
        """to be implemented in child class"""
        pass


class CPUNormalizer(Normalizer):
    """Normalizer cleans and standardizes text input using regular expression
    substitutions. Lowercasing is also applied if desired.

    Normalizer specific for CPU computation.

    Attributes:
        fields : Union[str, List[str]]
            the name of the column(s) containing text to be normalized
        regex_set : List[RegexSubPair]
            a list of dicionaries with keys, 'regex_str' and 'sub_str' used to
            perform regular expression substitutions of the text
        lowercase : bool
            whether or not the text will be lowercased

    Methods:
        run : run the step's primary function
        get_parameters : get a dictionary representation of the step object
        set_write_config : Alter the default file writing configuration
        write_outputs : Write output data to a file
    """
    def __init__(self, *args, **kwargs):
        """
        Keyword Arguments:
            fields : Union[str, List[str]]
                the name of the column(s) containing text to be normalized
            regex_set : List[RegexSubPair]
                a list of dicionaries with keys, 'regex_str' and 'sub_str' used to
                perform regular expression substitutions of the text
            lowercase : bool
                whether or not the text will be lowercased
        """
        super().__init__(*args, **kwargs)
        # Rust uses a different syntax for regex match groups
        # than standard (uses '$' instead of backslash)
        for regex_sub_pair in self.regex_set:
            regex_sub_pair['sub_str'] = regex_sub_pair['sub_str'].replace("\\", "$")

    def run(self, data: pa.Table,
            artifacts: dict = None) -> Tuple[pa.Table, dict]:
        """Runs a CPU-based normalizer method based on the configuration
        used to create the object of the CPUNormalizer class

        Arguments:
            data : pyarrow.Table
                a pyarrow Table containing at least one text column of type
                string or large_string
            artifacts : dict
                artifacts are not used in this run method, but must be received
                to operate correctly in the pipeline run method

        Returns:
            pyarrow.Table of cleaned data
        """
        # Perform validations
        validate_pyarrow_table(data=data)
        validate_str_cols(fields=self.fields, data=data)

        def implement_regex_substitutions(df: pl.DataFrame) -> pl.DataFrame:
            """Reusable function to apply each regex substitution
            normalization to each string field"""

            for regex_sub_pair in self.regex_set:
                df = df.with_columns(
                    [pl.col(field).str.replace_all(
                        pattern=regex_sub_pair['regex_str'],
                        value=regex_sub_pair['sub_str']) for field in self.fields])
            return df

        # Use the Polars library to apply the normalization methods
        # to each field of the Table that is specified in self.fields
        df = (pl.from_arrow(data)
              .with_columns([
                  pl.col(field).str.to_lowercase()
                  for field in self.fields if self.lowercase])
              .pipe(implement_regex_substitutions))

        data = df.to_arrow()

        return (data, None)



"""Curated set of regular expressions for cleaning text from pathology reports"""

from typing import List

from bardi.nlp_engineering.regex_library import regex_lib
from bardi.nlp_engineering.regex_library.regex_set import RegexSet, RegexSubPair


class PathologyReportRegexSet(RegexSet):
    def __init__(
        self,
        convert_escape_codes: bool = True,  # 0
        handle_whitespaces: bool = True,  # 1
        remove_urls: bool = True,  # 2
        remove_special_punct: bool = True,  # 3
        remove_multiple_punct: bool = True,  # 4
        handle_angle_brackets: bool = True,  # 5
        replace_percent_sign: bool = True,  # 6
        handle_leading_digit_punct: bool = True,  # 7
        remove_leading_punct: bool = True,  # 8
        remove_trailing_punct: bool = True,  # 9
        handle_words_with_punct_spacing: bool = True,  # 10
        handle_math_spacing: bool = True,  # 11
        handle_dimension_spacing: bool = True,  # 12
        handle_measure_spacing: bool = True,  # 13
        handle_cassettes_spacing: bool = True,  # 14
        handle_dash_digit_spacing: bool = True,  # 15
        handle_literals_floats_spacing: bool = True,  # 16
        fix_pluralization: bool = True,  # 17
        handle_digits_words_spacing: bool = True,  # 18
        remove_phone_numbers: bool = True,  # 19
        remove_dates: bool = True,  # 20
        remove_times: bool = True,  # 21
        remove_addresses: bool = True,  # 22
        remove_dimensions: bool = True,  # 23
        remove_specimen: bool = True,  # 24
        remove_decimal_seg_numbers: bool = True,  # 25
        remove_large_digits_seq: bool = True,  # 26
        remove_large_floats_seq: bool = True,  # 27
        trunc_decimals: bool = True,  # 28
        remove_cassette_names: bool = True,  # 29
        remove_duration_time: bool = True,  # 30
        remove_letter_num_seq: bool = True,
    ):
        # === List of regex sub pairs ===
        self.regex_set: List[RegexSubPair] = []

        # === Retrieve regular expression substitution pairs from regex_lib
        #  0 removes escapes codes
        if convert_escape_codes:
            self.regex_sub_escape_codes = regex_lib.get_escape_code_regex()
            self.regex_set.append(self.regex_sub_escape_codes)

        #  1 (regex, sub_str) for any new line, carriage return tab
        # and multiple spaces --> " "
        if handle_whitespaces:
            self.regex_sub_whitespaces = regex_lib.get_whitespace_regex()
            self.regex_set.append(self.regex_sub_whitespaces)

        #  2 removes URLs that start with https http or www
        if remove_urls:
            self.regex_sub_urls = regex_lib.get_urls_regex()
            self.regex_set.append(self.regex_sub_urls)

        #  3 matches a set of special punctuation
        # ,();[]#{}* --> " "
        if remove_special_punct:
            self.regex_sub_special_punct = regex_lib.get_special_punct_regex()
            self.regex_set.append(self.regex_sub_special_punct)

        #  4 matches multiple occurences of symbols like -, .and _
        if remove_multiple_punct:
            self.regex_sub_multiple_punct = regex_lib.get_multiple_punct_regex()
            self.regex_set.append(self.regex_sub_multiple_punct)

        #  5(regex, sub_str) removes angle brackets
        # <THIS IS INSIDE> --> THIS IS INSIDE
        if handle_angle_brackets:
            self.regex_sub_angle_brackets = regex_lib.get_angle_brackets_regex()
            self.regex_set.append(self.regex_sub_angle_brackets)

        #  6 (regex, sub_str) replaces % for a percent word 56% --> 56 PERCENT
        if replace_percent_sign:
            self.regex_sub_percent_sign = regex_lib.get_percent_sign_regex()
            self.regex_set.append(self.regex_sub_percent_sign)

        #  7
        if handle_leading_digit_punct:
            self.regex_sub_leading_digit_punct = (
                regex_lib.get_leading_digit_punctuation_regex()
            )
            self.regex_set.append(self.regex_sub_leading_digit_punct)

        #  8
        if remove_leading_punct:
            self.regex_sub_leading_punct = regex_lib.get_leading_punctuation_regex()
            self.regex_set.append(self.regex_sub_leading_punct)

        #  9
        if remove_trailing_punct:
            self.regex_sub_trailing_punct = regex_lib.get_trailing_punctuation_regex()
            self.regex_set.append(self.regex_sub_trailing_punct)

        #  10
        if handle_words_with_punct_spacing:
            self.regex_sub_words_with_punct_spacing = (
                regex_lib.get_words_with_punct_spacing_regex()
            )
            self.regex_set.append(self.regex_sub_words_with_punct_spacing)

        #  11
        if handle_math_spacing:
            self.regex_sub_math_spacing = regex_lib.get_math_spacing_regex()
            self.regex_set.append(self.regex_sub_math_spacing)

        #  12
        if handle_dimension_spacing:
            self.regex_sub_dimension_spacing = regex_lib.get_dimension_spacing_regex()
            self.regex_set.append(self.regex_sub_dimension_spacing)

        #  13
        if handle_measure_spacing:
            self.regex_sub_measure_spacing = regex_lib.get_measure_spacing_regex()
            self.regex_set.append(self.regex_sub_measure_spacing)

        #  14
        if handle_cassettes_spacing:
            self.regex_sub_cassette_spacing = regex_lib.get_cassettes_spacing_regex()
            self.regex_set.append(self.regex_sub_cassette_spacing)

        #  15
        if handle_dash_digit_spacing:
            self.regex_sub_dash_spacing = regex_lib.get_dash_digits_spacing_regex()
            self.regex_set.append(self.regex_sub_dash_spacing)

        #  16
        if handle_literals_floats_spacing:
            self.regex_sub_literals_floats_spacing = (
                regex_lib.get_literals_floats_spacing_regex()
            )
            self.regex_set.append(self.regex_sub_literals_floats_spacing)

        #  17
        if fix_pluralization:
            self.regex_sub_fix_pluralization = regex_lib.get_fix_pluralization_regex()
            self.regex_set.append(self.regex_sub_fix_pluralization)

        #  18
        if handle_digits_words_spacing:
            self.regex_sub_digits_words_spacing = (
                regex_lib.get_digits_words_spacing_regex()
            )
            self.regex_set.append(self.regex_sub_digits_words_spacing)

        #  19 remove phone numbers
        if remove_phone_numbers:
            self.regex_sub_phone_numbers = regex_lib.get_phone_number_regex()
            self.regex_set.append(self.regex_sub_phone_numbers)

        #  20 remove dates and times
        if remove_dates:
            self.regex_sub_dates = regex_lib.get_dates_regex()
            self.regex_set.append(self.regex_sub_dates)

        #  21 remove time
        if remove_times:
            self.regex_sub_time = regex_lib.get_time_regex()
            self.regex_set.append(self.regex_sub_time)

        #  22 remove addresses
        if remove_addresses:
            self.regex_sub_address = regex_lib.get_address_regex()
            self.regex_set.append(self.regex_sub_address)

        #  23 dimension substitution
        if remove_dimensions:
            self.regex_sub_dimensions = regex_lib.get_dimensions_regex()
            self.regex_set.append(self.regex_sub_dimensions)

        #  24 specimen
        if remove_specimen:
            self.regex_sub_specimen = regex_lib.get_specimen_regex()
            self.regex_set.append(self.regex_sub_specimen)

        #  25
        if remove_decimal_seg_numbers:
            self.regex_sub_decimal_seg_numbers = (
                regex_lib.get_decimal_segmented_numbers_regex()
            )
            self.regex_set.append(self.regex_sub_decimal_seg_numbers)

        #  26
        if remove_large_digits_seq:
            self.regex_sub_large_digits_seq = regex_lib.get_large_digits_seq_regex()
            self.regex_set.append(self.regex_sub_large_digits_seq)

        #  27
        if remove_large_floats_seq:
            self.regex_sub_large_floats_seq = regex_lib.get_large_float_seq_regex()
            self.regex_set.append(self.regex_sub_large_floats_seq)

        #  28
        if trunc_decimals:
            self.regex_sub_trunc_decimals = regex_lib.get_trunc_decimals_regex()
            self.regex_set.append(self.regex_sub_trunc_decimals)

        #  29
        if remove_cassette_names:
            self.regex_sub_cassette_names = regex_lib.get_cassette_name_regex()
            self.regex_set.append(self.regex_sub_cassette_names)

        # 30
        if remove_duration_time:
            self.regex_sub_duration_time = regex_lib.get_duration_regex()
            self.regex_set.append(self.regex_sub_duration_time)

        # 31
        if remove_letter_num_seq:
            self.regex_sub_letter_num_seq = regex_lib.get_letter_num_seq_regex()
            self.regex_set.append(self.regex_sub_letter_num_seq)

        #  LAST condense spacing - not configurable, always executes
        self.regex_sub_spaces = regex_lib.get_spaces_regex()
        self.regex_set.append(self.regex_sub_spaces)



"""Segment the dataset into splits, such as test, train, and val"""

from abc import abstractmethod
from typing import Dict, List, NamedTuple, Tuple, Union

import numpy as np
import polars as pl
import pyarrow as pa

from bardi.nlp_engineering.utils.validations import validate_pyarrow_table
from bardi.pipeline import DataWriteConfig, Step

MapSplit = NamedTuple(
    "MapSplit", [("unique_record_cols", List[str]), ("split_mapping", Dict[str, str])]
)
MapSplit.__doc__ = (
    "Specify the requirements for splitting data exactly in line with "
    "an existing data split"
)
MapSplit.unique_record_cols.__doc__ = (
    "List of column names of which the combination forms a unique identifier in the "
    "dataset. This can be a single column name if that creates a unique identifier, "
    "but oftentimes in datasets a combination of fields are required. "
    "Note: This set of columns MUST create a unique record or the program will crash. "
)
MapSplit.split_mapping.__doc__ = (
    "A dictionary mapping where the keys are the hash of the concatenated values from "
    "unique_record_cols, or represented by the following pseudocode, "
    "hash(concat(*unique_record_cols)). The values are thecorresponding split value "
    "(train, test, val)."
)

NewSplit = NamedTuple(
    "NewSplit",
    [
        ("split_proportions", Dict[str, float]),
        ("unique_record_cols", List[str]),
        ("group_cols", List[str]),
        ("label_cols", List[str]),
        ("random_seed", int),
    ],
)
NewSplit.__doc__ = (
    "Specify the requirements for splitting data with a new split from scratch."
)
NewSplit.split_proportions.__doc__ = (
    "Mapping of split names to split proportions. "
    "i.e. {'train': 0.75, 'test': 0.15, 'val': 0.15} "
    "Note: values must add to 1.0."
)
NewSplit.unique_record_cols.__doc__ = (
    "List of column names of which the combination forms a unique identifier in the "
    "dataset. This can be a single column name if that creates a unique identifier, "
    "but oftentimes in datasets a combination of fields are required. "
    "Note: This set of columns MUST create a unique record or the program will crash. "
)
NewSplit.group_cols.__doc__ = (
    "List of column names that form a 'group' that you would like to keep in discrete "
    "splits. E.x., if you had multiple medical notes for a single patient, "
    "you may desire that all notes for a single patient end up in the same split "
    "to prevent potential information leakage. In this case you would provide "
    "something like a patient_id."
)
NewSplit.label_cols.__doc__ = (
    "List of column names containing labels. Efforts are made to balance label "
    "distribution across splits, but this is not guaranteed."
)
NewSplit.random_seed.__doc__ = (
    "Required for reproducibility. If you have no preference, try on 42 for size."
)


class Splitter(Step):
    """Parent class to hardware specific child classes.
    Do not instantiate this parent class directly."""

    def __init__(
        self,
        split_method: Union[MapSplit, NewSplit],
    ):
        """Create an object of the Splitter class

        Keyword Arguments:
            split_method : Union[MapSplit, NewSplit]
                A named tuple of either MapSplit type or NewSplit
                type. Each contains a different set of values
                used to create the splitter depending upon split type
        """
        if isinstance(split_method, MapSplit):
            self.split_type = "map"
            self.unique_record_cols = split_method.unique_record_cols
            self.split_mapping = split_method.split_mapping
        elif isinstance(split_method, NewSplit):
            self.split_type = "new"
            self.split_proportions = split_method.split_proportions
            self.num_splits = len(self.split_proportions)
            self.unique_record_cols = split_method.unique_record_cols
            self.group_cols = split_method.group_cols
            self.label_cols = split_method.label_cols
            self.random_seed = split_method.random_seed
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }  # Default write config

    @abstractmethod
    def run(self):
        """to be implemented in child class"""
        pass


class CPUSplitter(Splitter):
    """The splitter adds a 'split' column to the data assigning
    each record to a particular split for downstream model training.

    Two split types are available - creating a new random split
    from scratch and assigning previously created splits. This
    second option is helpful when running comparisons with
    other methods of data processing ensuring that splits are
    exactly the same.

    Attributes:
        split_type : str
        unique_record_cols : List[str]
        split_mapping : dict[str, str]
        split_proportions : dict[str, float]
        num_splits : int
        group_cols : List[str]
        label_cols : List[str]
        random_seed: int

    Methods:
        run : run the step's primary function
        get_parameters : get a dictionary representation of the step object
        set_write_config : Alter the default file writing configuration
        write_outputs : Write output data to a file
    """

    def __init__(self, *args, **kwargs):
        """Create an object of the Splitter class to be run on a single-node CPU.

        Keyword Arguments:
            split_method : Union[MapSplit, NewSplit]
                A named tuple of either MapSplit type or NewSplit
                type. Each contains a different set of values
                used to create the splitter depending upon split type.
                Import these from bardi.nlp_engineering.splitter.
        """
        super().__init__(*args, **kwargs)

    def run(self, data: pa.Table, artifacts: dict = None) -> Tuple[pa.Table, dict]:
        """Runs a splitter using CPU computation based on the configuration
        used to create the object of the CPUSplitter class

        Keyword Arguments:
            data : PyArrow Table
                The data to be split
            artifacts : dict
                artifacts are not consumed in this run method, but must be
                received to operate correctly in the pipeline run method

        Returns:
            Tuple(PyArrow Table, dict)
                A tuple with:
                 - the first element holding a PyArrow Table of data
                 including the new split column
                 - the second element of the tuple intended for
                 artifacts is None

        Raises
            TypeError
                The run method was not supplied a PyArrow Table
        """
        # Perform validations
        validate_pyarrow_table(data=data)

        df = pl.from_arrow(data)

        # Mapping from an existing data split - Good for comparisons
        if self.split_type == "map":
            # Apply an existing mapping of id -> split

            # Combine and hash the columns that form a unique/distinct
            # record and map a dictionary set up as:
            # {id: split} where id is also a combined and hashed
            # combination of the columns forming a unique record
            df = (
                df.with_columns(
                    [
                        pl.concat_str([*self.unique_record_cols])
                        .hash()
                        .alias("composite_record_id")
                    ]
                )
                .with_columns(
                    [
                        pl.col("composite_record_id")
                        .cast(pl.Utf8())
                        .map_dict(self.split_mapping)
                        .alias("split")
                    ]
                )
                .drop("composite_record_id")
            )

        elif self.split_type == "new":
            # Create a new split of the data if an existing desired
            # split does not yet exist

            # Get the set of distinct groups from the overall data
            distinct_groups = df.select([*self.group_cols]).unique(maintain_order=True)
            group_count = distinct_groups.height
            # Generate a permutation of indices of distinct groups.
            np.random.seed(self.random_seed)
            permuted_indices = np.random.permutation(group_count)

            # The splits assignment has to be done with a number
            # so a mapping must be created from a number to the
            # desired split name
            split_name_maps = {
                i: key for i, key in enumerate(self.split_proportions.keys())
            }

            # Create a numpy array to hold the split labels (digits)
            split_labels = np.empty(group_count, dtype=int)
            start, end = 0, 0
            # Loop over the split proportion and assign digit
            # labels to the corresponding indices. If 15% of
            # data should be in a given split then we take a next
            # slice of permutated indices of that size and assign to them
            # a given digit label.
            for i, split_proportion in enumerate(self.split_proportions.items()):
                end = start + int(split_proportion[1] * group_count)
                split_labels[permuted_indices[start:end]] = i
                start = end
            # In case some of incides are left out due to divisibility.
            split_labels[permuted_indices[end:]] = i
            split_labels = pl.Series(split_labels)

            # Add split labels as a column to distince groups
            # mapped to assigned value back to the name of
            # the split
            splits = distinct_groups.with_columns(
                split_labels.map_dict(split_name_maps).alias("split")
            )

            # Join the split set of distinct groups back to the rest
            # of the data. If a group had more than one record in the
            # dataset, this join will ensure those records are mapped
            # to the same split as the others in the group
            df = df.join(splits, on=[*self.group_cols], how="left")

        data = df.to_arrow()

        return (data, None)



"""Define RegexSet and RegexSubPair blueprints"""

from typing import List, TypedDict


class RegexSubPair(TypedDict):
    """Dictionary used for regular expression string substitutions

    Attributes:
        regex_str: regular expression pattern
        sub_str: replacement value for matched string

    Example:
        {"regex_str": r"\\s",
         "sub_str": "WHITESPACE"}
    """

    regex_str: str
    sub_str: str


class RegexSet:
    """Blueprint for creating a configurable, domain specific regular expression set"""

    def __init__(self):
        self.regex_set: List[RegexSubPair] = []

    def get_regex_set(self) -> List[RegexSubPair]:
        """Return the ordered set of regular expressions"""
        return self.regex_set
